"""
Classes and functions to read WW files.
"""

from dataclasses import dataclass
from pathlib import Path
from typing import Any, Dict, List, TextIO

import numpy as np
from numpy import float64
from numpy.typing import NDArray

from f4enix.input.ww_gvr.models import NestedList, Vectors
from f4enix.input.ww_gvr.utils import compose_b2_vectors
from f4enix.output.meshtal import Fmesh, Meshtal


@dataclass()
class WWHeader:
    if_: int  # File type. Always 1, unused.
    iv: int  # Time-dependent windows flag. 1/2 = no/yes
    ni: int  # Number of particle types
    nr: int  # 10/16/16 = cartesian/cylindrical/spherical coord
    probid: str  # Date and time of run
    ne: List[int]  # Number of energy windows of each particle
    nfx: int  # Number of fine meshses in i
    nfy: int  # Number of fine meshses in j
    nfz: int  # Number of fine meshses in k
    origin: List[float]  #  Bottom left corner for cart, bottom center for cyl
    ncx: int  # Number of coarse meshses in i
    ncy: int  # Number of coarse meshses in j
    ncz: int  # Number of coarse meshses in k


@dataclass()
class WWHeaderCyl(WWHeader):
    director_1: List[float]  # Vector defining the director 1
    director_2: List[float]  # Vector defining the director 2


@dataclass()
class ParseResult:
    header: WWHeader
    b2_vectors: Vectors
    energies: NestedList
    values: NestedList


def parse(path: Path) -> ParseResult:
    """
    Parse a WW file and return a ParseResult object containing all the information of
    the file.

    Parameters
    ----------
    path : Path
        Path to the WW file.

    Returns
    -------
    ParseResult
        A ParseResult object containing all the information of the file.
    """
    with open(path) as infile:
        header = _parse_header(infile)

        b2_vectors = Vectors(
            vector_i=_read_block_2_vector(infile, header.ncx),
            vector_j=_read_block_2_vector(infile, header.ncy),
            vector_k=_read_block_2_vector(infile, header.ncz),
        )

        energies, values = _read_block_3(infile, header)

    return ParseResult(
        header=header, b2_vectors=b2_vectors, energies=energies, values=values
    )


def _parse_header(infile: TextIO) -> WWHeader:
    words = infile.readline().split()
    if_ = int(words[0])
    iv = int(words[1])
    ni = int(words[2])
    nr = int(words[3])
    probid = " ".join(words[4:])

    words = infile.readline().split()
    ne = [int(word) for word in words]  # Number of energy windows of each particle

    words = infile.readline().split()
    nfx = int(float(words[0]))  # Number of fine meshes in i
    nfy = int(float(words[1]))  # Number of fine meshes in j
    nfz = int(float(words[2]))  # Number of fine meshes in k
    origin = [float(word) for word in words[3:]]

    words = infile.readline().split()
    ncx = int(float(words[0]))  # Number of coarse meshes in i
    ncy = int(float(words[1]))  # Number of coarse meshes in j
    ncz = int(float(words[2]))  # Number of coarse meshes in k

    header_args: Dict[str, Any] = {
        "if_": if_,
        "iv": iv,
        "ni": ni,
        "nr": nr,
        "probid": probid,
        "ne": ne,
        "nfx": nfx,
        "nfy": nfy,
        "nfz": nfz,
        "origin": origin,
        "ncx": ncx,
        "ncy": ncy,
        "ncz": ncz,
    }

    if nr == 10:  # Cartesian coordinates
        header = WWHeader(**header_args)
    else:  # Cylindrical coordinates
        director_1 = [float(word) for word in words[3:]]
        words = infile.readline().split()
        director_2 = [float(word) for word in words[:3]]
        header_args.update(
            {
                "director_1": director_1,
                "director_2": director_2,
            }
        )
        header = WWHeaderCyl(**header_args)

    return header


def _read_block_2_vector(
    infile: TextIO, number_of_coarse_intervals: int
) -> NDArray[float64]:
    """
    Block 2 vectors combine the info of the coarse and fine meshes, they give the
    position of each coarse mesh point and the amount of fine intervals between them as:
    [mesh_position, fine_ints,
     mesh_position, 1.0000, fine_ints,
     mesh_position, 1.0000, fine_ints,
     ...]
    """
    expected_length = 3 * number_of_coarse_intervals + 1
    vector: List[float] = []
    while len(vector) < expected_length:
        words = infile.readline().split()
        vector.extend([float(word) for word in words])

    return np.array(vector)


def _read_block_3(infile: TextIO, header: WWHeader) -> tuple[NestedList, NestedList]:
    energies: NestedList = []
    values: NestedList = []

    for particle_index in range(header.ni):
        # Read energy bins for this particle
        expected_energy_bins = header.ne[particle_index]
        energies_current_particle: List[float] = []
        while len(energies_current_particle) < expected_energy_bins:
            words = infile.readline().split()
            energies_current_particle.extend([float(word) for word in words])

        # Read the values for this particle, all the energy bins at once
        expected_value_bins = (
            header.nfx * header.nfy * header.nfz * expected_energy_bins
        )
        values_current_particle: List[float] = []
        while len(values_current_particle) < expected_value_bins:
            words = infile.readline().split()
            values_current_particle.extend([float(word) for word in words])

        # Populate the result lists
        energies.append(energies_current_particle)
        values.append(values_current_particle)

    return energies, values


def read_meshtally_file(path: Path, tally_id: int = 0) -> ParseResult:
    meshtal = Meshtal(path)
    meshtal.readMesh()

    # If no tally id is provided, select the first meshtally
    if tally_id == 0:
        tally_id = next(iter(meshtal.mesh.keys()))

    mesh = meshtal.mesh[tally_id]

    header = _read_header_from_meshtally_file(mesh)

    b2_vectors = _calculate_b2_vectors_from_meshtally(mesh)

    energies = [[100.0]]  # Only one particle and one energy bin
    values = [list(mesh.dat.flatten())]

    return ParseResult(
        header=header,
        b2_vectors=b2_vectors,
        energies=energies,
        values=values,
    )


def _read_header_from_meshtally_file(mesh: Fmesh) -> WWHeader:
    if_ = 1  # It is always 1
    iv = 1  # It is always 1, time-dependent windows not supported
    ni = 1  # Number of particle types, always 1 when reading a meshtally
    nr = 10 if mesh.cart else 16
    ne = [1]  # Number of energy bins of each particle, always 1 for a meshtally

    # When reading a meshtally we consider that there are no regular intervals between
    # coarse vectors. This means that the length of a coarse vector is equal to the
    # total amount of fine ints at that dimension
    nfx = ncx = len(mesh.dims[3]) - 1
    nfy = ncy = len(mesh.dims[2]) - 1
    nfz = ncz = len(mesh.dims[1]) - 1
    origin = [mesh.origin[3], mesh.origin[2], mesh.origin[1]]

    header_args: Dict[str, Any] = {
        "if_": if_,
        "iv": iv,
        "ni": ni,
        "nr": nr,
        "probid": "Generated by f4enix_ww",
        "ne": ne,
        "nfx": nfx,
        "nfy": nfy,
        "nfz": nfz,
        "origin": origin,
        "ncx": ncx,
        "ncy": ncy,
        "ncz": ncz,
    }

    if nr == 10:  # Cartesian coordinates
        header = WWHeader(**header_args)
    else:  # Cylindrical coordinates
        director_1 = mesh.axis.tolist()
        if mesh.vec is None:  # MCNP5 Meshtally have no info about vec
            if director_1[0] != 1:
                director_2 = [1.0, 0.0, 0.0]  # Default MCNP vec
            else:
                director_2 = [0.0, 1.0, 0.0]  # Cant be the same as the axis
        else:
            director_2 = mesh.vec.tolist()
        header_args.update(
            {
                "director_1": director_1,
                "director_2": director_2,
            }
        )
        header = WWHeaderCyl(**header_args)

    return header


def _calculate_b2_vectors_from_meshtally(mesh: Fmesh) -> Vectors:
    if mesh.cart:  # In cartesian coordinates the vectors start at the origin
        coarse_vectors = Vectors(
            vector_i=np.array(mesh.dims[3]) + np.array(mesh.origin[3]),
            vector_j=np.array(mesh.dims[2]) + np.array(mesh.origin[2]),
            vector_k=np.array(mesh.dims[1]) + np.array(mesh.origin[1]),
        )
    else:  # In cylindrical the vectors are not modified by the origin
        coarse_vectors = Vectors(
            vector_i=np.array(mesh.dims[3]),
            vector_j=np.array(mesh.dims[2]),
            vector_k=np.array(mesh.dims[1]),
        )

    fine_vectors = Vectors(
        vector_i=np.ones(len(mesh.dims[3])),
        vector_j=np.ones(len(mesh.dims[2])),
        vector_k=np.ones(len(mesh.dims[1])),
    )

    return compose_b2_vectors(coarse_vectors, fine_vectors)


def write(
    file_path: Path,
    header: WWHeader,
    b2_vectors: Vectors,
    energies: NestedList,
    values: NestedList,
) -> None:
    """
    Write a WW file from the given information.

    Parameters
    ----------
    file_path : Path
        Path to the file to be written.
    header : WWHeader
        The header information of the WW file.
    b2_vectors : Vectors
        The block 2 vectors of the WW file.
    energies : NestedList
        The energies of the WW file.
    values : NestedList
        The values of the WW file.

    Returns
    -------
    None
    """
    with open(file_path, "w") as infile:
        _write_header(infile, header)
        _write_block_2(infile, b2_vectors)
        _write_block_3(infile, energies, values)


def _write_header(f: TextIO, header: WWHeader) -> None:
    f.write(f"{header.if_:>10}{header.iv:>10}{header.ni:>10}{header.nr:>10}")
    f.write(f"{header.probid:>38}\n")

    for e_bins in header.ne:
        f.write(f"{e_bins:>10}")
    f.write("\n")

    f.write(f"{header.nfx:>#9.5g}{header.nfy:>#13.5g}{header.nfz:>#13.5g}")
    f.write(f"{header.origin[0]:>#13.5g}")
    f.write(f"{header.origin[1]:>#13.5g}")
    f.write(f"{header.origin[2]:>#13.5g}\n")

    f.write(f"{header.ncx:>#9.5g}{header.ncy:>#13.5g}{header.ncz:>#13.5g}")
    if type(header) == WWHeaderCyl:
        f.write(f"{header.director_1[0]:>#13.5g}")
        f.write(f"{header.director_1[1]:>#13.5g}")
        f.write(f"{header.director_1[2]:>#13.5g}\n")
        f.write(f"{header.director_2[0]:>#9.5g}")
        f.write(f"{header.director_2[1]:>#13.5g}")
        f.write(f"{header.director_2[2]:>#13.5g}")
        f.write(f"{2:>#13.5g}")
    else:
        f.write(f"{1:>#13.5g}")
    f.write("\n")


def _write_block_2(f: TextIO, b2_vectors: Vectors) -> None:
    for vector in b2_vectors:
        vector = vector.tolist()
        packs_of_six = [vector[i : i + 6] for i in range(0, len(vector), 6)]

        for pack in packs_of_six:
            first_number = pack.pop(0)
            f.write(f"{first_number:>#9.5g}")
            for number in pack:
                f.write(f"{number:>#13.5g}")
            f.write("\n")


def _write_block_3(f: TextIO, energies: NestedList, values: NestedList) -> None:
    for particle_index in range(len(energies)):
        particle_ergs = energies[particle_index]
        packs_of_6 = [particle_ergs[i : i + 6] for i in range(0, len(particle_ergs), 6)]
        for energy_pack in packs_of_6:
            first_energy = energy_pack.pop(0)
            f.write(f"{first_energy:>#9.5g}")
            for energy in energy_pack:
                f.write(f"{energy:>#13.5g}")
            f.write("\n")

        particle_vals = values[particle_index]
        packs_val = [particle_vals[i : i + 6] for i in range(0, len(particle_vals), 6)]
        for value_pack in packs_val:
            for value in value_pack:
                f.write(f"{value:>#13.4E}")
            f.write("\n")



"""
This file includes the main class of the ww_gvr package, the WW class.
"""
# flake8: noqa: PLR2004
from pathlib import Path
from typing import List, Optional

import numpy as np

from f4enix.input.ww_gvr import ww_parser
from f4enix.input.ww_gvr.geometry import Geometry
from f4enix.input.ww_gvr.models import (
    CoordinateType,
    EnergiesByParticle,
    NestedList,
    ParticleType,
    Pathlike,
    ValuesByParticle,
)
from f4enix.input.ww_gvr.ratios_calculation import calculate_max_ratio_array
from f4enix.input.ww_gvr.utils import decompose_b2_vectors
from f4enix.input.ww_gvr.ww_parser import ParseResult, WWHeader, WWHeaderCyl


class WW:
    def __init__(
        self,
        file_path: Path,
        geometry: Geometry,
        values: ValuesByParticle,
    ):
        """
        The representation of a WW.

        It should be created using the class methods WW.load_from_ww_file or
        WW.create_gvr_from_meshtally_file.

        Attributes
        ----------
        file_path : Path
            Path to the MCNP weight window file.
        geometry : Geometry
            A Geometry object that contains the mesh and all geometrical information.
        values : ValuesByParticle
            A dictionary with the values of the weight window for each combination of
            particle and energy.
        """
        self.file_path = file_path
        self.geometry = geometry
        self.values = values

    @classmethod
    def load_from_ww_file(cls, file_path: Pathlike):
        """
        Create a WW object from a MCNP weight window file.

        Parameters
        ----------
        file_path : Pathlike
            Path to the MCNP weight window file.

        Returns
        -------
        WW
            A WW object.
        """
        file_path = Path(file_path)
        parse_result = ww_parser.parse(file_path)

        geometry = cls._create_geometry(parse_result)
        values = cls._nested_to_values_by_particle(parse_result)

        return WW(file_path, geometry, values)

    @classmethod
    def create_gvr_from_meshtally_file(
        cls,
        file_path: Pathlike,
        maximum_splitting_ratio: float = 5.0,
        softening_factor: float = 1.0,
    ):
        """
        Create a WW object (a GVR) from a MCNP meshtally file using the Van Vick /
        Andrew Davis / Magic algorithm.

        Parameters
        ----------
        file_path : Pathlike
            Path to the MCNP meshtally file.
        maximum_splitting_ratio : float, optional
            The maximum splitting ratio used in the formula, by default 5.0

        Returns
        -------
        WW
            A WW object.
        """
        file_path = Path(file_path)
        parse_result = ww_parser.read_meshtally_file(file_path)

        geometry = cls._create_geometry(parse_result)
        values = cls._nested_to_values_by_particle(parse_result)
        cls._convert_flux_values_to_gvr(values, maximum_splitting_ratio)
        gvr = WW(file_path, geometry, values)
        
        if softening_factor != 1.0:
            gvr.soften(softening_factor)
        
        return gvr

    @staticmethod
    def _create_geometry(parse_result: ParseResult) -> Geometry:
        coarse_vectors, fine_vectors = decompose_b2_vectors(parse_result.b2_vectors)
        return Geometry(parse_result.header, coarse_vectors, fine_vectors)

    @staticmethod
    def _nested_to_values_by_particle(parse_result: ParseResult) -> ValuesByParticle:
        particles = [ParticleType.NEUTRON]
        if parse_result.header.ni > 1:
            particles.append(ParticleType.PHOTON)

        values_by_particle: ValuesByParticle = {}
        for i, particle in enumerate(particles):
            values_by_particle[particle] = {}
            energies = parse_result.energies[i]
            reshaped_values = np.array(parse_result.values[i]).reshape(
                len(energies),
                parse_result.header.nfz,
                parse_result.header.nfy,
                parse_result.header.nfx,
            )

            for j, energy in enumerate(energies):
                values_by_particle[particle][energy] = reshaped_values[j]

        return values_by_particle

    def _format_nested_energies_and_values(self) -> tuple[NestedList, NestedList]:
        """The opposite of _nested_to_values_by_particle. Returns energies and values
        in the format that the writer of ww_parser expects."""
        energies = []
        values = []

        for particle in self.particles:
            particle_energies = list(self.values[particle].keys())
            particle_energies = np.array(particle_energies).flatten().tolist()
            energies.append(particle_energies)

            particle_values = list(self.values[particle].values())
            particle_values = np.array(particle_values).flatten().tolist()
            values.append(particle_values)

        return energies, values

    def _calculate_ratios(self) -> ValuesByParticle:
        ratios_by_particle = {}

        for particle in self.particles:
            ratios_by_particle[particle] = {}

            for energy in self.energies[particle]:
                ratios_by_particle[particle][energy] = calculate_max_ratio_array(
                    self.values[particle][energy]
                )

        return ratios_by_particle

    @staticmethod
    def _convert_flux_values_to_gvr(
        values: ValuesByParticle, maximum_splitting_ratio: float
    ) -> None:
        # Only one particle is expected when reading a meshtally file
        all_values = np.concatenate(list(values[ParticleType.NEUTRON].values()))
        max_flux = np.max(all_values)

        for energy in values[ParticleType.NEUTRON].keys():
            # Van Vick / Andrew Davis / Magic algorithm
            values[ParticleType.NEUTRON][energy] /= max_flux
            values[ParticleType.NEUTRON][energy] *= 2 / (maximum_splitting_ratio + 1)

    @property
    def values(self) -> ValuesByParticle:
        """
        The values of the weight window for each combination of particle and energy as
        a nested dictionary.

        Has the following format: {ParticleType: {energy(float): np.ndarray}}
        """
        return self._values

    @values.setter
    def values(self, values: ValuesByParticle):
        self._values = values
        self._ratios = self._calculate_ratios()
        self.geometry.fill_grid_values(self._values)
        self.geometry.fill_grid_ratios(self._ratios)

    @property
    def ratios(self) -> ValuesByParticle:
        """
        Same format as values, but showing the maximum ratio of every voxel with its
        neighbours.
        """
        return self._ratios

    @property
    def energies(self) -> EnergiesByParticle:
        """
        The energies of the weight window for each particle as a dictionary.

        Has the following format: {ParticleType: List[float]}
        """
        energies = {}
        for particle in self.particles:
            energies[particle] = list(self._values[particle].keys())
        return energies

    @property
    def particles(self) -> List[ParticleType]:
        """
        List of ParticleType objects that are present in the weight window.

        It will be either [ParticleType.NEUTRON] or [ParticleType.NEUTRON,
        ParticleType.PHOTON].
        """
        particles = [ParticleType.NEUTRON]
        if len(self.values) > 1:
            particles.append(ParticleType.PHOTON)
        return particles

    @property
    def info(self) -> str:
        """
        A string with the most important information of the weight window.
        """
        text = f" {self.file_path.name} weight window\n"

        vi = self.geometry._coarse_vectors.vector_i
        vj = self.geometry._coarse_vectors.vector_j
        vk = self.geometry._coarse_vectors.vector_k
        text += f"       {'From':-^12} {'To':-^12} {'No. Bins':-^12}\n"
        text += f" I --> {vi[0]:^12.1f} {vi[-1]:^12.1f} {self.geometry.i_ints:^12.0f}\n"
        text += f" J --> {vj[0]:^12.1f} {vj[-1]:^12.1f} {self.geometry.j_ints:^12.0f}\n"
        text += f" K --> {vk[0]:^12.1f} {vk[-1]:^12.1f} {self.geometry.k_ints:^12.0f}\n"

        coordinates = self.geometry.coordinate_type.name.capitalize()
        no_voxels = self.geometry.i_ints * self.geometry.j_ints * self.geometry.k_ints
        text += f"\n {coordinates} coordinates, {no_voxels} voxels.\n"

        text += f"\n The weight window contains {len(self.particles)} particle/s"
        for particle in self.particles:
            text += f"\n\n {particle.name.capitalize():-^50}\n"

            all_values = np.concatenate(list(self.values[particle].values()))
            min_value = np.min(all_values)
            max_value = np.max(all_values)
            amount_positive_values = int(np.sum(all_values > 0))
            percent_positive_values = amount_positive_values / all_values.size * 100
            all_ratios = np.concatenate(list(self.ratios[particle].values()))
            average_ratio = np.average(all_ratios)
            max_ratio = np.max(all_ratios)

            text += f" {'Energy bins:':<20} {self.energies[particle]}"
            text += f"\n {'Min value:':<20} {min_value:.2E}"
            text += f"\n {'Max value:':<20} {max_value:.2E}"
            text += f"\n {'No.Bins > 0 [%]:':<20} {percent_positive_values:.1f}%"
            text += f"\n {'Average ratio:':<20} {average_ratio:.2E}"
            text += f"\n {'Max ratio:':<20} {max_ratio:.2E}"

        return text

    def __repr__(self) -> str:
        return self.info

    def multiply(self, factor: float) -> None:
        """
        Multiply all values of the weight window by a factor.

        Parameters
        ----------
        factor : float
            The factor to multiply the values.

        Returns
        -------
        None
        """
        if factor == 1.0:
            return

        values = self.values
        for particle in self.particles:
            for energy in self.energies[particle]:
                values[particle][energy] *= factor

        # This line will trigger the setter, which will update the ratios and geometry
        self.values = values

    def soften(self, softening_factor: float) -> None:
        """
        Soften the weight window values by raising them to a power.

        This is useful to mitigate long histories and to reduce the aggressiveness of
        the weight window if the softening factor is lower than 1. Softening factors
        higher than 1 will have the opposite effect.

        Parameters
        ----------
        softening_factor : float
            The softening factor.

        Returns
        -------
        None
        """
        if softening_factor == 1.0:
            return

        values = self.values
        for particle in self.particles:
            for energy in self.energies[particle]:
                values[particle][energy] **= softening_factor

        # This line will trigger the setter, which will update the ratios and geometry
        self.values = values

    def add_particle(self, norm: float = 1, soft: float = 1):
        """
        Add a second particle to a weight window that only has one (neutrons).

        The new particle will be photons, and its values will be the same as the
        neutron values but multiplied by a normalization factor and softened.

        Parameters
        ----------
        norm : float, optional
            The normalization factor, by default 1
        soft : float, optional
            The softening factor, by default 1

        Returns
        -------
        None
        """
        if len(self.particles) == 2:
            raise ValueError("There are already two particles...")

        values_photon = {
            energy: value * norm**soft
            for energy, value in self.values[ParticleType.NEUTRON].items()
        }

        values = self.values
        values.update({ParticleType.PHOTON: values_photon})

        # This line will trigger the setter, which will update the ratios and geometry
        self.values = values

    def remove_particle(self):
        """
        Remove a particle from a weight window that has two (neutrons and photons).

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        if len(self.particles) == 1:
            raise ValueError("There is only one particle already...")

        self.values = {ParticleType.NEUTRON: self.values[ParticleType.NEUTRON]}

    def mitigate_long_histories(self, max_ratio: float = 10.0) -> None:
        """
        The weigth window voxels with a ratio higher than max_ratio will be set to
        zero. This stops the particle from over-splitting.

        Parameters
        ----------
        max_ratio : float, optional
            The maximum ratio allowed, by default 10.0

        Returns
        -------
        None
        """
        values = self.values
        for particle in self.particles:
            for energy in self.energies[particle]:
                mask_high_ratios = self.ratios[particle][energy] > max_ratio
                values[particle][energy][mask_high_ratios] = 0.0

        # This line will trigger the setter, which will update the ratios and geometry
        self.values = values

    def write_to_ww_file(self, file_path: Optional[Pathlike] = None) -> None:
        """
        Write the weight window to a file.

        Parameters
        ----------
        file_path : Optional[Pathlike], optional
            The path to the file, by default None. If None, the file will be written
            with the same name as the original file but with "_written" added to the
            name.

        Returns
        -------
        None
        """
        if file_path is None:
            file_path = self.file_path.parent / (self.file_path.stem + "_written")
        file_path = Path(file_path)

        nested_energies, nested_values = self._format_nested_energies_and_values()
        ww_parser.write(
            file_path,
            self._build_header(),
            self.geometry.b2_vectors,
            nested_energies,
            nested_values,
        )

    def _build_header(self) -> WWHeader:
        header_args = {
            "if_": 1,
            "iv": 1,
            "ni": len(self.particles),
            "nr": 10
            if self.geometry.coordinate_type == CoordinateType.CARTESIAN
            else 16,
            "probid": "Created by f4enix_ww",
            "ne": [len(energies) for energies in self.energies.values()],
            "nfx": self.geometry.i_ints,
            "nfy": self.geometry.j_ints,
            "nfz": self.geometry.k_ints,
            "origin": self.geometry.origin,
            "ncx": self.geometry.i_coarse_ints,
            "ncy": self.geometry.j_coarse_ints,
            "ncz": self.geometry.k_coarse_ints,
        }
        if self.geometry.coordinate_type == CoordinateType.CARTESIAN:
            header = WWHeader(**header_args)
        else:  # CoordinateType.CYLINDRICAL
            header_args["director_1"] = self.geometry.director_1
            header_args["director_2"] = self.geometry.director_2
            header = WWHeaderCyl(**header_args)

        return header

    def export_as_vtk(self, file_path: Optional[Pathlike] = None) -> None:
        """
        Export the weight window to a VTK file.

        Parameters
        ----------
        file_path : Optional[Pathlike], optional
            The path to the file, by default None. If None, the file will be written
            with the same name as the original file but with ".vtk" extension.

        Returns
        -------
        None
        """
        if file_path is None:
            file_path = self.file_path.parent / (self.file_path.stem)
        file_path = Path(file_path)

        self.geometry.export_as_vtk(file_path)

    def plot(self) -> None:
        """
        Plot the weight window using the pyvista plotter.

        The plot is interactive, so it will block the execution of the script until the
        plot is closed.

        Parameters
        ----------
        None

        Returns
        -------
        None
        """
        self.geometry.plot()



from __future__ import annotations

"""
Classes and types used throughout the package.
"""


from dataclasses import dataclass
from enum import Enum
from pathlib import Path
from typing import Dict, List, Union

import numpy as np
from numpy.typing import NDArray


class CoordinateType(Enum):
    CARTESIAN = "cartesian"
    CYLINDRICAL = "cylindrical"


class ParticleType(Enum):
    NEUTRON = "neutron"
    PHOTON = "photon"


@dataclass()
class Vectors:
    """
    A group of three vectors in the form of a Numpy NDArray. They are used to define 3D
    meshes.
    """

    vector_i: NDArray  # Vector defining i direction
    vector_j: NDArray  # Vector defining j direction
    vector_k: NDArray  # Vector defining k direction

    def __iter__(self):
        return iter([self.vector_i, self.vector_j, self.vector_k])

    def __eq__(self, other):
        try:
            np.testing.assert_array_almost_equal(self.vector_i, other.vector_i)
            np.testing.assert_array_almost_equal(self.vector_j, other.vector_j)
            np.testing.assert_array_almost_equal(self.vector_k, other.vector_k)
            return True
        except AssertionError:
            return False


Pathlike = Union[str, Path]
NestedList = List[List[float]]
ValuesByEnergy = Dict[float, np.ndarray]
ValuesByParticle = Dict[ParticleType, ValuesByEnergy]
EnergiesByParticle = Dict[ParticleType, List[float]]



from __future__ import annotations
from typing import IO, Dict, Tuple, Union

import numpy as np
import pyvista as pv

from f4enix.input.ww_gvr.models import Pathlike


def get_volume_sampling_error_from_cuv(
    file_path: Pathlike, voxel_sampling_points: int
) -> Dict[int, Dict[int, float]]:
    """
    Read the first mesh tally from a CUV file and return the volume sampling error
    """
    with open(file_path) as infile:
        line = infile.readline()
        while line:
            if "INDEX" in line:
                return _read_cuv_data(infile, voxel_sampling_points)

            line = infile.readline()
        return {}


def _read_cuv_data(
    infile: IO, voxel_sampling_points: int
) -> Dict[int, Dict[int, float]]:
    voxel_cell_errors = {}

    while True:
        try:
            voxel_id, cell_errors = _read_voxel(infile, voxel_sampling_points)
            voxel_cell_errors[voxel_id] = cell_errors
        except IndexError:
            break

    return voxel_cell_errors


def _read_voxel(infile: IO, voxel_sampling_points: int) -> Tuple[int, Dict[int, float]]:
    first_line_words = infile.readline().split()
    # 1-based index on the file, we want 0-based
    voxel_id = int(first_line_words[0]) - 1
    amount_of_cells = int(first_line_words[5])

    cell_errors = {}
    for _ in range(amount_of_cells):
        words = infile.readline().split()
        cell_id = int(words[0])
        cell_error = calculate_volume_sampling_error(
            float(words[1]), voxel_sampling_points
        )
        cell_errors[cell_id] = cell_error

    # no need to read the CUV values and relative errors
    for _ in range(2 * amount_of_cells):
        infile.readline()

    return voxel_id, cell_errors


def calculate_volume_sampling_error(
    partial_volume: float, voxel_sampling_points: int
) -> float:
    return (
        (partial_volume - partial_volume**2) / (voxel_sampling_points)
    ) ** 0.5 / partial_volume


pyvista_grid = Union[
    pv.StructuredGrid, pv.UnstructuredGrid, pv.PolyData, pv.RectilinearGrid
]


def add_sampling_error_to_vtk(
    grid: pyvista_grid, cuv_file_path: Pathlike, voxel_sampling_points: int
) -> pyvista_grid:
    """
    Add a new array to the grid with the volume sampling error. In the case of a voxel
    with more than one cell, the error is the biggest one.
    """
    voxel_cell_errors = get_volume_sampling_error_from_cuv(
        cuv_file_path, voxel_sampling_points
    )

    vector_errors = np.zeros(grid.n_cells)
    for i in range(grid.n_cells):
        cell_error = max(voxel_cell_errors.get(i, {}).values(), default=0)
        vector_errors[i] = cell_error

    grid["Volume Sampling Error"] = vector_errors
    return grid



from pathlib import Path

import pytest
import pyvista as pv

from f4enix.output.cuv_sampling_error import (
    add_sampling_error_to_vtk,
    calculate_volume_sampling_error,
    get_volume_sampling_error_from_cuv,
)
from f4enix.output.meshtal import Meshtal

PATH_TO_CUV_FILE = (
    Path(__file__).parent / "resources" / "meshtal" / "test_CuV" / "cuvmsh"
)
ASSUMED_VOXEL_SAMPLING_POINTS = 1000


def test_get_volume_samplig_error_from_cuv():
    voxel_cell_errors = get_volume_sampling_error_from_cuv(
        file_path=PATH_TO_CUV_FILE, voxel_sampling_points=ASSUMED_VOXEL_SAMPLING_POINTS
    )

    assert voxel_cell_errors[1][1] == pytest.approx(0.0)
    assert voxel_cell_errors[6][1] == pytest.approx(0.01918304084569366)
    assert voxel_cell_errors[6][10] == pytest.approx(0.05212937865502626)


@pytest.mark.parametrize(
    ("partial_volume", "voxel_sampling_points", "expected_result"),
    [
        (1.0, 1000, 0.0),
        (0.9, 1000, 0.010540925533894595),
        (0.5, 1000, 0.03162277660168379),
        (0.5, 10, 0.31622776601683794),
    ],
)
def test_calculate_volume_sampling_error(
    partial_volume, voxel_sampling_points, expected_result
):
    result = calculate_volume_sampling_error(partial_volume, voxel_sampling_points)
    assert result == pytest.approx(expected_result)


def test_add_sampling_error_to_vtk(tmpdir):
    # Read a meshtal and create a VTK file
    meshtal = Meshtal(PATH_TO_CUV_FILE)
    meshtal.readMesh()
    meshtal.mesh[44].write(tmpdir)
    grid = pv.read(tmpdir / "cuvmsh_44_vtk.vtr")
    assert type(grid) == pv.RectilinearGrid

    grid_with_errors = add_sampling_error_to_vtk(
        grid=grid, cuv_file_path=PATH_TO_CUV_FILE, voxel_sampling_points=1000
    )

    assert grid_with_errors["Volume Sampling Error"][0] == pytest.approx(0.0)  # type: ignore
    assert grid_with_errors["Volume Sampling Error"][5] == pytest.approx(  # type: ignore
        0.05121707142933009
    )



"""
This file contains functions to create and operate pyvista meshes of cartesian and
cylindrical grids.

These meshes follow the convention of k, j, i vectors.
    j
    --- ---
    | 3 | 4 |
    --- ---
    | 1 | 2 |
    --- ---    i	 

The values are given in an array of shape (k, j, i).
"""

from typing import Optional

import numpy as np
import pyvista as pv
from numpy.typing import NDArray


def create_cartesian_grid(
    vector_i: NDArray,
    vector_j: NDArray,
    vector_k: NDArray,
    origin: Optional[NDArray] = None,
) -> pv.StructuredGrid:
    """
    Create a cartesian grid with the given vectors and origin.

    Parameters
    ----------
    vector_i : NDArray
        Vector defining the i direction intervals.
    vector_j : NDArray
        Vector defining the j direction intervals.
    vector_k : NDArray
        Vector defining the k direction intervals.
    origin : NDArray | None
        Origin vector of the grid, by default None (0,0,0).
    """
    # meshgrid should receive float arrays to avoid a warning
    vector_i = np.asarray(vector_i, "float32")
    vector_j = np.asarray(vector_j, "float32")
    vector_k = np.asarray(vector_k, "float32")

    x_coordinates, y_coordinates, z_coordinates = np.meshgrid(
        vector_i, vector_j, vector_k, indexing="ij"
    )
    grid = pv.StructuredGrid(x_coordinates, y_coordinates, z_coordinates)

    if origin is not None:
        grid.translate(origin, inplace=True)

    return grid


def create_cylindrical_grid(
    vector_i: NDArray,
    vector_j: NDArray,
    vector_k_revolutions: NDArray,
    origin=None,
    axis=None,
    vec=None,
) -> pv.StructuredGrid:
    """
    Create a cylindrical grid with the given vectors and origin.

    In the case of a single theta interval or two theta intervals, the theta vector is
    artificially extended to 20 intervals as they couldnt be visually represented.

    Parameters
    ----------
    vector_i : NDArray
        Vector defining the radial direction intervals.
    vector_j : NDArray
        Vector defining height direction intervals.
    vector_k_revolutions : NDArray
        Vector defining k direction intervals in revolution units.
    origin : NDArray | None
        Vector pointing to the bottom center of the cylinder, by default None (0,0,0).
    axis : NDArray | None
        Vector defining the axis of the cylinder, by default None (Z-axis).
    vec : NDArray | None
        Vector defining the direction of the radius, by default None (X-axis).
    """
    vector_i = _correct_radius_vector(vector_i)
    vector_k_revolutions = _correct_theta_vector(vector_k_revolutions)

    # Extend the theta vector if less than 3 intervals
    if len(vector_k_revolutions) <= 3:
        vector_k_revolutions = _extend_theta_intervals(vector_k_revolutions)

    # Create the grid on the Z axis
    grid = _create_cylindrical_grid_with_z_axis(
        vector_i, vector_j, vector_k_revolutions
    )

    # Roto translate the cylinder
    if vec is not None:
        _rotate_to_vec(grid, vec)
    if axis is not None:
        _rotate_to_axis(grid, axis)
    if origin is not None:
        grid.translate(origin, inplace=True)

    return grid


def _extend_theta_intervals(vector_k: NDArray, new_theta_ints: int = 20):
    if len(vector_k) == 2:
        extended_vector_k = np.linspace(vector_k[0], vector_k[1], new_theta_ints + 1)

    elif len(vector_k) == 3:
        new_theta_ints = new_theta_ints // 2
        vector_k_0 = np.linspace(vector_k[0], vector_k[1], new_theta_ints + 1)
        vector_k_1 = np.linspace(vector_k[1], vector_k[2], new_theta_ints + 1)[1:]
        extended_vector_k = np.concatenate((vector_k_0, vector_k_1))

    else:
        raise ValueError("Only 1 or 2 theta defined intervals can be extended...")

    return extended_vector_k


def _create_cylindrical_grid_with_z_axis(
    vector_i: NDArray, vector_j: NDArray, vector_k_revolutions: NDArray
) -> pv.StructuredGrid:
    vector_k_radians = np.array(vector_k_revolutions) * 2 * np.pi

    radii, thetas, z = np.array(
        np.meshgrid(vector_i, vector_k_radians, vector_j, indexing="ij")
    )
    x = np.cos(thetas) * radii
    y = np.sin(thetas) * radii
    points = np.c_[x.ravel(order="f"), y.ravel(order="f"), z.ravel(order="f")]

    grid = pv.StructuredGrid()
    grid.points = points
    grid.dimensions = len(vector_i), len(vector_k_revolutions), len(vector_j)

    return grid


def _rotate_to_vec(cylinder: pv.StructuredGrid, vec: NDArray):
    """Rotate a Z axis cylinder by its axis to match the vec."""
    z_axis = np.array([0.0, 0.0, 1.0])
    default_vec = np.array([1.0, 0.0, 0.0])

    # Calculate the angle to rotate
    angle = _angle_between(default_vec, vec)

    # Apply the rotation
    cylinder.rotate_vector(vector=z_axis, angle=angle, inplace=True)


def _rotate_to_axis(cylinder: pv.StructuredGrid, axis: NDArray):
    default_axis = np.array([0.0, 0.0, 1.0])
    normal_to_rotation = np.cross(default_axis, axis)

    # Calculate the angle to rotate
    angle = _angle_between(default_axis, axis)

    # Apply the rotation
    cylinder.rotate_vector(vector=normal_to_rotation, angle=angle, inplace=True)


def _correct_radius_vector(vector_i: NDArray):
    """A radius that start at zero generates degenerate cells at the center."""
    if np.isclose(vector_i[0], 0.0):
        vector_i[0] = vector_i[1] * 0.01
    return vector_i


def _correct_theta_vector(vector_k_revolutions: NDArray) -> NDArray:
    """With floats sometimes the first and last values are not exactly 0 and 1."""
    # If the cylinder starts at exactly 0 revolutions
    if np.isclose(vector_k_revolutions[0], 0.0):
        vector_k_revolutions[0] = 0.0

    # If the cylinder ends at exactly 1 revolution
    if np.isclose(vector_k_revolutions[-1], 1.0):
        vector_k_revolutions[-1] = 1.0

    return vector_k_revolutions


def _angle_between(vec_1: NDArray, vec_2: NDArray) -> float:
    """Returns the angle in degrees between vectors 'v1' and 'v2'"""
    v1_normalized = _normalize_vector(vec_1)
    v2_normalized = _normalize_vector(vec_2)
    radians = np.arccos(np.clip(np.dot(v1_normalized, v2_normalized), -1.0, 1.0))
    return np.rad2deg(radians)


def _normalize_vector(vector: NDArray) -> NDArray:
    """Returns the unit vector of the vector."""
    return vector / np.linalg.norm(vector)



import os
import math
import copy
import logging
import pandas as pd
import numpy as np

from numjuggler.parser import Card
from f4enix.input.MCNPinput import Input
from f4enix.constants import (
    SECTOR_BOUNDARIES,
    SECTOR_BOUNDARIES_ANGLES,
    SECTOR_NAMES,
    PLASMA_CELLS,
)


class Elite_Input(Input):
    def __init__(
        self,
        cells: list[Card],
        surfs: list[Card],
        data: list[Card],
        header: list = None,
    ) -> None:
        """Children of the :py:class:`Input`.

        It features the methods for sectors extraction from ITER E-Lite 360
        degree model

        Parameters
        ----------
        cells : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP cells.
        surfs : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP surfaces.
        data : list[parser.Card]
            list of numjuggler.parser.Card objects containing MCNP data cards.
        header : list, optional
            list of strings that compose the MCNP header, by default None

        Examples
        --------
        Read the input and the associated Excel file.

        By calling extract_sector you'll be able to get a working MCNP input
        file of the desired sector.

        >>> from f4enix.input.elite import Elite_Input
        ... elite_inp = Elite_Input.from_input('Elite.i')
        ... elite_inp.extract_sector(1, 'Elite_excel.xlsx', 'sector1.i', tol=1e-5, True)
        """
        super().__init__(cells, surfs, data, header)
        self.__initialized = False
        self._block_structure = None
        self._sectors_L0_cells_names = {}

    def _initialize_elite(self, excel_file: os.PathLike, check_Elite: bool = True):
        # checks if the input is actually e-lite and initializes some variables
        # check is optional
        self._block_structure = pd.read_excel(excel_file)

        # check the envelopes
        if check_Elite:
            inp_cells = self.get_cells_summary()
            inp_L0_cells = inp_cells[inp_cells["universe"].isna()].index.tolist()
            if set(inp_L0_cells) != set(self._block_structure["Cell"].tolist()):
                raise RuntimeError(
                    "MCNP input is not an E-Lite file, or the Excel Block"
                    + " Structure file is not compatible"
                )

        for sec in SECTOR_NAMES:
            # you will have to modify outer cell to put them in the sectors
            self._sectors_L0_cells_names[sec] = self._block_structure[
                self._block_structure["Sector"] == sec
            ]["Cell"].tolist()

        self.__initialized = True

    def extract_sector(
        self,
        sectors,
        excel_file: os.PathLike,
        outfile: os.PathLike = "sector",
        tol: float = 1e-5,
        check_Elite: bool = False,
    ) -> None:
        """Writes a working input of a user-selected E-Lite sector.

        The user can provide a sector number or a list of contiguous sector
        numbers in counterclockwise direction (e.g. 1, "2 & 3", [4,5], [9, 1]).
        Currently there is no check of the correctness of the input, the user
        must be careful in providing the correct sector number(s), in the
        correct order.
        The method will extract a working input of the selected sector(s), by
        replacing the boundaries with reflecting surfaces.
        The method is based on an auxiliary Excel file that lists all E-Lite
        envelope containers and their respective sector. The method follows
        these steps:

        - Excel file initialization and model check. If there's no correspondence
          between the envelope structure of E-Lite model and excel file, the
          method aborts. The method can't check if the correct sector number is
          assigned to the envelope container.

        - Collection of all envelope containers and fillers to be extracted

        - Graveyard and outer cell modification

        - Source term fixing

        - L0 surfaces which are enough close to the boundary will be set as reflective

        - L1 surfaces which are too close to the boundary will be translated
          "outwards" with respect to the sector(s), to avoid the arising of
          fatal errors

        Parameters
        ----------
        sectors : int or list
            sector number, or list of contiguous sector numbers in
            counterclockwise direction that will be extracted
        excel_file : os.PathLike
            path to the Excel file that describes the sector structure of the
            version of E-Lite in use
        outfile : os.PathLike, optional
            name of the input file that will be written, by default 'sector'
        tol : int, optional
            determines the maximum distance for two planes to be considered
            equal, and determines the magnitude of the translation of L1 cells.
            It should be chosen basedon the value used in DBCN card, by default 1e-5
        check_Elite : bool, optional
            Automatically checks if the envelope structure of the model is
            consistent with the one reported in the Excel, by default False
        """
        if not self.__initialized:
            self._initialize_elite(excel_file, check_Elite)

        # build list of sectors to be extracted
        if not isinstance(sectors, list):
            sectors = [sectors]

        logging.info("Collecting the cells, surfaces, materials and transf.")
        # collect L0 cells to be extracted
        cells = []
        for sector in sectors:
            cells += self._sectors_L0_cells_names[sector]
        # collect L0 surfaces, needed for later
        L0_cells = set(cells)
        L0_sset = set()
        for _, cell in enumerate(self.cells.values()):
            if cell.values[0][0] in L0_cells:
                for v, t in cell.values:
                    if t == "sur":
                        L0_sset.add(v)
        # backup copy graveyard and outercell, that will be modified
        # append gy and outercell manually as they don't belong to a sector
        cells.append(800)
        cells.append(801)
        # get cells, surfaces and materials to be extracted
        cells_cards, surf_dic, materials = self._extraction_function(
            cells, None, True, True
        )
        # copy the surfaces, because they will be modified
        modified_surfaces = copy.deepcopy(surf_dic)

        # Also tallies and other data
        modified_data_cards = copy.deepcopy(self.other_data)
        # pattern_str = '|'.join(self.tally_cards_types)
        # pattern = re.compile(f'^({pattern_str})\d+$')
        # # for now remove all tally cards
        # for key in self.other_data.keys():

        #     # Explanation of the pattern:
        #     # ^           - Start of the string
        #     # (pattern)   - A group containing possible patterns
        #     # \d+         - One or more digits
        #     # $           - End of the string
        #     if pattern.match(key):
        #         modified_data_cards.pop(key)
        # modify sdef
        Elite_Input._set_sdef(sectors, modified_data_cards)
        # set L0 as periodic and modify L1 planes
        Elite_Input._set_boundaries(
            Elite_Input._get_boundaries_angles(sectors),
            tol,
            L0_sset,
            modified_surfaces,
            self.transformations,
        )
        # modify graveyard and outercell
        new_outercell, new_gy = Elite_Input._modify_graveyard(
            sectors, self.cells["801"], self.cells["800"]
        )
        cells_cards["800"] = new_outercell
        cells_cards["801"] = new_gy
        # extract tallies based on comments
        # self._extract_tallies(sectors)
        # write final MCNP input
        Input.write_blocks(
            outfile,
            False,
            cells_cards,
            modified_surfaces,
            materials,
            self.header,
            self.transformations,
            modified_data_cards,
        )
        logging.info("input written correctly")

    @staticmethod
    def _set_sdef(sectors, modified_data_cards):
        # write new SI and SD cards, directly in input attribute
        # the copies are modified, original input is preserved
        new_si = "SI70 L "
        new_sp = "SP70 "
        for sector in sectors:
            new_si = new_si + str(PLASMA_CELLS[sector]) + " "
            if sector != "2 & 3":
                new_sp = new_sp + "1 "
            else:
                new_sp = new_sp + "2 "
        modified_data_cards["SI70"].input = [new_si]
        modified_data_cards["SP70"].input = [new_sp]
        return

    @staticmethod
    def _set_boundaries(
        boundaries_angles, tol, L0_sset, modified_surfaces, transformations
    ):
        # define the angles at which there are the planes cutting the sectors
        boundary_angles = []
        for angle in boundaries_angles:
            if angle > 180:
                rev_angle = angle - 180
            else:
                rev_angle = angle + 180
            boundary_angles.append(angle)
            boundary_angles.append(rev_angle)

        for l, angle in enumerate(boundary_angles):
            coeffs = np.array(
                [-math.sin(math.radians(angle)), math.cos(math.radians(angle)), 0]
            )
            # Check all surfaces for each angle
            for surf in modified_surfaces.values():
                # check if the surface belongs to L0 or L1 of the sector(s)
                if surf.values[0][0] in L0_sset:
                    bound_opt = 1
                else:
                    bound_opt = 2
                # check if the surface is a plane
                if surf.stype in ["p", "px", "py", "pz"]:
                    if surf.stype == "p":
                        p_coeffs = np.array(surf.scoefs[:3])
                        d = surf.scoefs[3]
                    elif surf.stype == "px":
                        p_coeffs = np.array([1, 0, 0])
                        d = surf.scoefs[0]
                    elif surf.stype == "py":
                        p_coeffs = np.array([0, 1, 0])
                        d = surf.scoefs[0]
                    elif surf.stype == "pz":
                        p_coeffs = np.array([0, 0, 1])
                        d = surf.scoefs[0]
                    p_coeffs = p_coeffs / np.linalg.norm(p_coeffs)
                    # check if the surface is a plane passing by the origin
                    if Elite_Input._check_tol(tol, [(d, 0)]):
                        # check if the plane has a transformation
                        try:
                            # if so, apply it and check if it's a plane parallel to
                            # the boundary
                            tr = surf._get_value_by_type("tr")
                            tr_name = "TR" + str(tr)
                            trans = transformations[tr_name]
                            # check if it's only a translation
                            if len(trans.values) < 12:
                                continue
                            # if it's a rotation, rotate the coefficients
                            p_coeffs = Elite_Input._rotate_plane(trans, p_coeffs)
                        # if not, just check if it's parallel to the boundary
                        except UnboundLocalError:
                            pass
                        # if the plane is similar to boundary planes, modify it
                        if Elite_Input._check_tol(tol, list(zip(p_coeffs, coeffs))):
                            Elite_Input._modify_boundary(surf, bound_opt, angle, l, tol)

    @staticmethod
    def _get_boundaries_angles(sectors):
        # get the angles of the two boundary surfaces, counterclockwise
        boundaries_angles = []

        for k, sec in enumerate(sectors):
            bounds = SECTOR_BOUNDARIES_ANGLES[sec]
            if k == 0:
                # get y- plane for first sector in list
                boundaries_angles.append(bounds[0])
            if k == (len(sectors) - 1):
                # get y+ plane for last sector in list
                boundaries_angles.append(bounds[1])

        return boundaries_angles

    @staticmethod
    def _modify_graveyard(sectors, graveyard, outercell):
        # cut/ union graveyard and outercell with planes
        for k, sector in enumerate(sectors):
            if k == 0:
                new_gy = Input.add_surface(
                    graveyard,
                    -SECTOR_BOUNDARIES[sector][0],
                    None,
                    "union",
                    False,
                )
                new_outercell = Input.add_surface(
                    outercell,
                    SECTOR_BOUNDARIES[sector][0],
                    None,
                    "intersect",
                    False,
                )
            if k == len(sectors) - 1:
                new_gy = Input.add_surface(
                    new_gy, -SECTOR_BOUNDARIES[sector][1], None, "union", False
                )
                new_outercell = Input.add_surface(
                    new_outercell,
                    SECTOR_BOUNDARIES[sector][1],
                    None,
                    "intersect",
                    False,
                )
            else:
                continue
        # put the newly created cells in F4Enix dict, they will be replaced
        return new_outercell, new_gy

    @staticmethod
    def _rotate_plane(trans, p_coeffs):
        # compute rotation matrix
        coeffs = [t[0] for t in trans.values]
        if trans.unit == "*":
            coeffs = [math.cos(math.radians(v)) for v in coeffs]
        # define rotation matrix
        rot_matrix = np.array(
            [
                [coeffs[4], coeffs[5], coeffs[6]],
                [coeffs[7], coeffs[8], coeffs[9]],
                [coeffs[10], coeffs[11], coeffs[12]],
            ]
        )
        # compute inverse rotation matrix
        inv_rot = np.linalg.inv(rot_matrix)
        # compute rotated plane's coefficients to be checked
        dp = np.dot(inv_rot, p_coeffs)

        return dp

    @staticmethod
    def _check_tol(tol, n_coeff):
        # for all tuples in the list, checks if the elements in the tuples
        # are within the tolerance
        in_tol = True
        # loop over the first list
        for coeff in n_coeff:
            if not coeff[1] - tol <= coeff[0] <= coeff[1] + tol:
                in_tol = False
                break
        return in_tol

    @staticmethod
    def _modify_boundary(surf, bound_opt, angle, l, tol):
        # if in L0, set periodic
        tol_sign = {True: 1, False: -1}
        if bound_opt == 1:
            surf.input[0] = "*" + surf.input[0]
        # if L1, translate outwards to avoid fatal errors
        elif bound_opt == 2:
            # only way is to modify 'lines' and recompute input and template
            surf_desc = []
            # skip comment lines
            for line in surf.lines:
                if not line.lower().startswith("c"):
                    surf_desc.append(line)
            # get plane coefficients
            words = " ".join(string.rstrip("\n") for string in surf_desc)
            words = words.split()
            # put the correct sign to the translation
            # check if y+ or y- boundary
            y_plus = l > 1
            # check in which quadrant the plane is
            angle_quadr = not (90 < angle < 270 or -270 < angle < -90)
            # check the sign of the y coefficient of the plane
            y_coeff = surf.scoefs[1] > 0
            # translate the plane and recompute template and input
            sign = tol_sign[y_plus] * tol_sign[angle_quadr] * tol_sign[y_coeff]
            words[-1] = "{:.8f}".format(float(words[-1]) + sign * 2 * tol)
            surf.lines = [" ".join(words) + "\n"]
            surf.get_input()



from __future__ import annotations

"""
This module contains the Geometry class, which is used to create a pyvista grid
from the coarse and fine vectors of a WW file. It also contains the methods to
fill the grid with the values of the WW file and to plot it.
"""

from pathlib import Path
from typing import Dict, List, Union

import numpy as np
import pyvista as pv
from numpy.typing import NDArray
from pyvista.plotting.plotter import Plotter

from f4enix.input.ww_gvr.meshgrids import create_cartesian_grid, create_cylindrical_grid
from f4enix.input.ww_gvr.models import CoordinateType, ValuesByParticle, Vectors
from f4enix.input.ww_gvr.utils import build_1d_vectors, compose_b2_vectors
from f4enix.input.ww_gvr.ww_parser import WWHeader, WWHeaderCyl


class Geometry:
    def __init__(
        self, header: WWHeader, coarse_vectors: Vectors, fine_vectors: Vectors
    ):
        """
        Class that contains all the geometrical information and methods associated to
        a WW file, including a pyvista mesh.

        Parameters
        ----------
        header : WWHeader
            The header of the WW file.
        coarse_vectors : Vectors
            The coarse vectors of the WW file.
        fine_vectors : Vectors
            The fine vectors of the WW file.
        """
        self._coarse_vectors = coarse_vectors
        self._fine_vectors = fine_vectors

        self._coordinate_type = (
            CoordinateType.CARTESIAN
            if type(header) == WWHeader
            else CoordinateType.CYLINDRICAL
        )

        self._director_1 = header.director_1 if type(header) == WWHeaderCyl else None
        self._director_2 = header.director_2 if type(header) == WWHeaderCyl else None
        self._origin = header.origin

        self._grid = self._create_grid()

    def _create_grid(self) -> pv.StructuredGrid:
        origin = np.array(self.origin)

        if self.coordinate_type == CoordinateType.CARTESIAN:
            return create_cartesian_grid(*self.vectors, origin=origin)

        else:
            vector_i, vector_j, vector_k = self.vectors
            axis = np.array(self.director_1)
            vec = np.array(self.director_2)
            return create_cylindrical_grid(
                vector_i, vector_j, vector_k, origin, axis, vec
            )

    def fill_grid_values(self, values_by_particle: ValuesByParticle) -> None:
        """
        Fills the grid with the values of the WW file.

        Parameters
        ----------
        values_by_particle : ValuesByParticle
            The values of the WW file grouped by particle.

        Returns
        -------
        None
        """
        for particle in values_by_particle.keys():
            for energy in values_by_particle[particle].keys():
                values = values_by_particle[particle][energy]
                self.fill_grid_array(values, f"WW {particle}, {energy:.2f} MeV")

    def fill_grid_ratios(self, ratios_by_particle: ValuesByParticle) -> None:
        """
        Fills the grid with the ratios of the WW file.

        Parameters
        ----------
        ratios_by_particle : ValuesByParticle
            The ratios of the WW file grouped by particle.

        Returns
        -------
        None
        """
        for particle in ratios_by_particle.keys():
            for energy in ratios_by_particle[particle].keys():
                ratios = ratios_by_particle[particle][energy]
                self.fill_grid_array(ratios, f"Ratio {particle}, {energy:.2f} MeV")

            max_ratios = np.max(list(ratios_by_particle[particle].values()), axis=0)
            self.fill_grid_array(max_ratios, f"Max Ratio {particle}")

    def fill_grid_array(self, array: NDArray, name: str) -> None:
        """
        Fills the grid with an array given in K, J, I order.

        If the grid is cylindrical, it properly manages the theta dimension even if it
        was extended with _extend_theta_intervals from the meshgrids module.

        Parameters
        ----------
        array : NDArray
            The array to fill the grid with.
        name : str
            The name of the array, key.
        """
        if self.coordinate_type == CoordinateType.CYLINDRICAL:
            array = self._prepare_values_for_cylindrical_grid(array)

        flattened_array = array.flatten()
        self._grid[name] = flattened_array

    def _prepare_values_for_cylindrical_grid(self, values: NDArray) -> NDArray:
        # If the grid was extended in the thetas, values should do the same
        total_ints = self.i_ints * self.j_ints * self.k_ints
        extended = self._grid.n_cells // total_ints
        if extended > 1:
            values = np.repeat(values, extended, axis=0)

        # The grids with cylindrical coordinates created with meshgrids.py
        #  have dimensions with the order (j, k, i) instead of (k, j, i)
        values = values.swapaxes(0, 1)

        return values

    @property
    def coordinate_type(self) -> CoordinateType:
        return self._coordinate_type

    @property
    def vectors(self) -> Vectors:
        """Returns a Vectors object (1D) with the coarse and fine vectors combined"""
        return build_1d_vectors(self._coarse_vectors, self._fine_vectors)

    @property
    def b2_vectors(self) -> Vectors:
        """Returns a Vectors object of the vectors in the b2 format of the WW file"""
        return compose_b2_vectors(self._coarse_vectors, self._fine_vectors)

    @property
    def director_1(self) -> Union[List[float], None]:
        """
        Returns the director 1 vector (axis) of the cylinder if the grid is
        cylindrical.
        """
        return self._director_1

    @property
    def director_2(self) -> Union[List[float], None]:
        """
        Returns the director 2 vector (radius) of the cylinder if the grid is
        cylindrical.
        """
        return self._director_2

    @property
    def origin(self) -> List[float]:
        return self._origin

    @property
    def i_ints(self) -> int:
        """Returns the number of intervals in the i direction"""
        return np.sum(self._fine_vectors.vector_i)

    @property
    def j_ints(self) -> int:
        """Returns the number of intervals in the j direction"""
        return np.sum(self._fine_vectors.vector_j)

    @property
    def k_ints(self) -> int:
        """Returns the number of intervals in the k direction"""
        return np.sum(self._fine_vectors.vector_k)

    @property
    def i_coarse_ints(self) -> int:
        """Returns the number of coarse intervals in the i direction"""
        return len(self._coarse_vectors.vector_i) - 1

    @property
    def j_coarse_ints(self) -> int:
        """Returns the number of coarse intervals in the j direction"""
        return len(self._coarse_vectors.vector_j) - 1

    @property
    def k_coarse_ints(self) -> int:
        """Returns the number of coarse intervals in the k direction"""
        return len(self._coarse_vectors.vector_k) - 1

    def plot(self) -> None:
        """Plots the grid with the values of the WW file in an interactive window"""
        args_dict = self._decide_plot_parameters()
        plotter = Plotter()
        plotter.add_mesh_clip_plane(self._grid, **args_dict)
        plotter.show_grid(color="black")  # type: ignore
        plotter.show_axes()  # type: ignore
        plotter.show()

    def _decide_plot_parameters(self) -> Dict:
        sclar_bar_args = {
            "interactive": True,
            "title_font_size": 20,
            "label_font_size": 16,
            "shadow": True,
            "italic": True,
            "fmt": "%.e",
            "font_family": "arial",
        }
        args_dict = {
            "scalars": self._grid.array_names[0],
            "scalar_bar_args": sclar_bar_args,
            "show_edges": False,
            "cmap": "jet",
            "log_scale": False,
            "lighting": False,
        }

        # Decide about the log scale
        data = np.array(self._grid[args_dict["scalars"]])
        data_max = data.max()
        data_min = data[data != 0].min()
        if data_max / data_min < 1e5:
            args_dict["log_scale"] = False
            return args_dict

        args_dict["log_scale"] = True
        exp_max = int(np.log10(data_max))
        data_max = 10**exp_max
        data_min = 10 ** (exp_max - 10)
        args_dict["clim"] = [data_min, data_max]
        args_dict["n_colors"] = 10
        args_dict["scalar_bar_args"]["n_labels"] = 10

        return args_dict

    def export_as_vtk(self, file_path: Path) -> None:
        """
        Exports the grid as a VTK file.

        Parameters
        ----------
        file_path : Path
            The path to create the VTK file.
        """
        if file_path.suffix != ".vts":
            file_path = file_path.with_suffix(".vts")

        self._grid.save(str(file_path))



"""Parse legacy FISPACT-II output files to extract the generic pathways.

"""

from __future__ import annotations

import re
import os
import pandas as pd
from dataclasses import dataclass
from f4enix.input.libmanager import LibManager


perc_pattern = re.compile(r"\d+\.*\d*%")
target_pathway_zaids = re.compile(r"[A-Z][a-z]*\s*\d+m*")
metastable_pat = re.compile(r"\d+m")
isotope_pat = re.compile(r"\d+")
element_pat = re.compile(r"[a-zA-Z]+")
path_id_pat = re.compile(r"\s+path")
reaction_pat = re.compile(r"\([a-zA-Z\d,+-]+\)")


@dataclass
class FispactZaid:
    """Class to represent a Zaid in a generic pathway in fispact.

    Attributes
    ----------
    element : str
        The element of the isotope.
    isotope : int
        The isotope number.
    metastable : bool
        Whether the isotope is metastable or not.
    """

    element: str
    isotope: int
    metastable: bool = False

    def get_str(self) -> str:
        """Return the Zaid as a string.

        Returns
        -------
        str
            The Zaid as a string.
        """
        if self.metastable:
            return f"{self.element}{self.isotope}m"
        else:
            return f"{self.element}{self.isotope}"


@dataclass
class Pathway:
    """Class to represent a pathway in fispact.

    Attributes
    ----------
    parent : FispactZaid
        The parent FispactZaid object.
    daughter : FispactZaid
        The daughter FispactZaid object.
    perc : float
        The percentage of the pathway contribution to the daughter isotope.
    reactions : list[str]
        The list of reactions in the pathway.
    intermediates : list[FispactZaid], optional
        The list of intermediate FispactZaid objects in the pathway.

    """

    parent: FispactZaid
    daughter: FispactZaid
    perc: float
    reactions: list[str]
    intermediates: list[FispactZaid] = None

    def __post_init__(self):
        # check that reaction list length is always one more than intermediates
        # when it is not None
        if self.intermediates is not None:
            try:
                assert len(self.reactions) == len(self.intermediates) + 1
            except AssertionError as e:
                raise AssertionError(
                    f"Number of reactions is too low for pathway {self}"
                ) from e

    def __str__(self) -> str:
        if self.intermediates is not None:
            text = f"{self.parent.get_str()} "
            for intermediate, reaction in zip(self.intermediates, self.reactions):
                text += f"-{reaction}-> {intermediate.get_str()} "
            text += f"-{self.reactions[-1]}->  {self.daughter.get_str()}"
            return text
        else:
            return f"{self.parent.get_str()} -{self.reactions[0]}-> {self.daughter.get_str()}"


class PathwayCollection:
    def __init__(self, pathways: list[Pathway]) -> None:
        """Collection of pathways. This can be created from a list of
        Pathway objects or directly from a FISPACT legacy output file.

        Parameters
        ----------
        pathways : list[Pathway]
            list of Pathway objects.

        Attributes
        ----------
        pathways : list[Pathway]
            list of Pathway objects.

        Examples
        --------
        >>> from f4enix.output.fispact_legacy_out import PathwayCollection
        ... collection = PathwayCollection.from_file("path/to/fispact/output")
        ... pathway = collection.pathways[0]
        ... print(pathway)
        ... print(pathway.parent, pathway.daughter, pathway.reactions,
        ...       patway.intermediates, pathway.perc)
        Mn55 -(n,g)-> Mn56
        Mn55 Mn56 ['n,g'] None 100.0

        """
        self.pathways = pathways

    @staticmethod
    def _get_zaid_from_str(name: str) -> FispactZaid:
        element = element_pat.search(name).group()
        isotope = isotope_pat.search(name).group()
        if metastable_pat.search(name) is not None:
            metastable = True
        else:
            metastable = False

        return FispactZaid(element=element, isotope=isotope, metastable=metastable)

    @classmethod
    def from_file(cls, file: os.PathLike) -> PathwayCollection:
        """
        Retrieve pathways from a FISPACT legacy output file and return a list of
        Pathway objects.

        Parameters
        ----------
        file : os.PathLike
            The path to the legacy FISPACT output file.

        Returns
        -------
        list[Pathway]
            A list of Pathway objects representing the pathways found in the file.
        """
        lines = []
        with open(file, "r", encoding="utf-8") as infile:
            for i, line in enumerate(infile):
                lines.append(line)
                if "Significant loops" in line:
                    start_reactions = i
                elif "G E N E R I C   P A T H W" in line:
                    end_reactions = i
        lines = lines[start_reactions + 1 : end_reactions]

        paths = []
        for i, line in enumerate(lines):
            if path_id_pat.match(line) is not None:
                perc = float(perc_pattern.search(line).group()[:-1])
                zaids = target_pathway_zaids.findall(line)
                try:
                    parent = cls._get_zaid_from_str(zaids[0])
                except IndexError as e:
                    print(line)
                    raise e
                daughter = cls._get_zaid_from_str(zaids[-1])
                if len(zaids) > 2:
                    intermediates = []
                    for zaid in zaids[1:-1]:
                        intermediates.append(cls._get_zaid_from_str(zaid))
                else:
                    intermediates = None

                # if a path is found, the following line will contain details
                # on the reactions
                line = lines[i + 1]
                reactions = reaction_pat.findall(line)

                try:
                    pathway = Pathway(
                        parent=parent,
                        daughter=daughter,
                        reactions=reactions,
                        intermediates=intermediates,
                        perc=perc,
                    )
                except AssertionError as e:
                    # Handle assertion error
                    print(lines[i])
                    print(line)
                    raise e

                paths.append(pathway)

        return PathwayCollection(paths)

    def to_dataframe(self) -> pd.DataFrame:
        """Return the pathways as a pandas DataFrame.
        columns are "Parent", "% contribution", "Intermediates", "Reactions",
        "Daughter".

        Returns
        -------
        pd.DataFrame
            The pathways as a pandas DataFrame.
        """
        # get a libmanager for the sorting
        lm = LibManager()

        rows = []
        for pathway in self.pathways:
            intermediates = []
            if pathway.intermediates is not None:
                for intermediate in pathway.intermediates:
                    intermediates.append(intermediate.get_str())
            rows.append(
                [
                    pathway.parent.get_str(),
                    intermediates,
                    pathway.reactions,
                    pathway.daughter.get_str(),
                    pathway.perc,
                    # additional columns for sorting to be dropped later
                    lm.get_zaidnum(pathway.daughter.get_str()),
                ]
            )

        df = pd.DataFrame(rows)
        df.columns = [
            "Parent",
            "Intermediates",
            "Reactions",
            "Daughter",
            "% contribution",
            "isotope_sort",
        ]
        df.sort_values(by=["isotope_sort"], inplace=True)
        df.set_index(["Daughter", "% contribution"], inplace=True)
        del df["isotope_sort"]
        return df



"""
Some utilities for the ww_gvr package.
"""

from pathlib import Path
from typing import Tuple

import numpy as np

from f4enix.input.ww_gvr.models import Vectors

PATH_TO_RESOURCES = Path(__file__).parents[3] / "tests" / "test_ww_gvr" / "resources"

WW_SIMPLE_CART = PATH_TO_RESOURCES / "ww_simple_cart"
WW_SIMPLE_CYL = PATH_TO_RESOURCES / "ww_simple_cyl"
WW_COMPLEX_CART = PATH_TO_RESOURCES / "ww_complex_cart"
MESHTALLY_CART = PATH_TO_RESOURCES / "meshtally_cart"
MESHTALLY_CYL = PATH_TO_RESOURCES / "meshtal_cyl"
EXPECTED_INFO_SIMPLE_CART = PATH_TO_RESOURCES / "expected_info_simple_cart.txt"


def decompose_b2_vectors(b2_vectors: Vectors) -> Tuple[Vectors, Vectors]:
    """
    Takes a Vectors object with b2 format and returns two Vectors objects, one with the
    coarse vectors and the other with the fine vectors.

    The b2 format of a WW MCNP file is:
            [mesh_position, fine_ints,
            mesh_position, 1.0000, fine_ints,
            mesh_position, 1.0000, fine_ints,
            ...
            mesh_position, 1.0000]

    Coarse vectors have the format: [p1, p2, p3, ...]
    A fine vector shows how many intervals there are between each step of the coarse
    vector's points: [n1, n2, n3, ...]

    Parameters
    ----------
    b2_vectors : Vectors
        Vectors object with b2 format.

    Returns
    -------
    Tuple[Vectors, Vectors]
        Two Vectors objects, one with the coarse vectors and the other with the
        fine vectors.
    """
    coarse = []
    fine = []

    for b2_vector in b2_vectors:
        b2_vector = np.insert(b2_vector, 1, 1.0000)
        coarse.append([b2_vector[i] for i in range(len(b2_vector)) if i % 3 == 0])
        fine.append(
            [
                int(b2_vector[j + 2])
                for j in range(len(b2_vector))
                if j % 3 == 0 and j + 2 < len(b2_vector)
            ]
        )

    coarse_vectors = Vectors(
        vector_i=np.array(coarse[0]),
        vector_j=np.array(coarse[1]),
        vector_k=np.array(coarse[2]),
    )
    fine_vectors = Vectors(
        vector_i=np.array(fine[0]),
        vector_j=np.array(fine[1]),
        vector_k=np.array(fine[2]),
    )

    return coarse_vectors, fine_vectors


def compose_b2_vectors(coarse_vectors: Vectors, fine_vectors: Vectors) -> Vectors:
    """
    The opposite of decompose_b2_vectors.

    Parameters
    ----------
    coarse_vectors : Vectors
        Vectors object with coarse vectors.
    fine_vectors : Vectors
        Vectors object with fine vectors.

    Returns
    -------
    Vectors
        Vectors object with b2 format.
    """
    b2_vectors = []

    for coarse_vector, fine_ints in zip(coarse_vectors, fine_vectors):
        current_vector = []
        for i in range(len(coarse_vector) - 1):
            current_vector.append(coarse_vector[i])
            current_vector.append(1.0000)
            current_vector.append(fine_ints[i])
        current_vector.append(coarse_vector[-1])
        current_vector.append(1)
        current_vector.remove(1)
        b2_vectors.append(current_vector)
    return Vectors(
        vector_i=np.array(b2_vectors[0]),
        vector_j=np.array(b2_vectors[1]),
        vector_k=np.array(b2_vectors[2]),
    )


def build_1d_vectors(coarse_vectors: Vectors, fine_vectors: Vectors) -> Vectors:
    """
    Flatten coarse and fine vectors into 1D Vectors.

    Takes two Vectors objects, one with the coarse vectors and the other with the
    fine vectors which represent the amount of ints between each step of the coarse
    vectors, and returns some flattened vectors with format:
    [mesh_position, mesh_position, mesh_position, ...]

    Parameters
    ----------
    coarse_vectors : Vectors
        Vectors object with coarse vectors.
    fine_vectors : Vectors
        Vectors object with fine vectors.

    Returns
    -------
    Vectors
        Vectors object with 1D vectors.
    """
    b2_vectors = []

    for coarse_vector, fine_ints in zip(coarse_vectors, fine_vectors):
        current_vector = [coarse_vector[0]]
        for i in range(len(coarse_vector) - 1):
            current_vector += np.linspace(
                coarse_vector[i], coarse_vector[i + 1], int(fine_ints[i] + 1)
            ).tolist()[1:]

        b2_vectors.append(current_vector)

    return Vectors(
        vector_i=np.array(b2_vectors[0]),
        vector_j=np.array(b2_vectors[1]),
        vector_k=np.array(b2_vectors[2]),
    )



"""
Includes some numba functions for improved performance of the ratios calculation.
"""

import numpy as np
from numba import njit, prange


@njit(cache=True)
def calculate_max_ratio_array(array: np.ndarray) -> np.ndarray:
    """
    Calculate the maximum ratio of each value to its neighbours in a 3D array.

    Parameters
    ----------
    array : np.ndarray
        The input 3D array.

    Returns
    -------
    np.ndarray
        A 3D array of the same shape as the input, where each value is the maximum
        ratio of the corresponding value in the input array to its neighbours.
    """
    ratios = np.ones_like(array)
    for k in prange(array.shape[0]):
        for j in prange(array.shape[1]):
            for i in prange(array.shape[2]):
                value = array[k, j, i]
                neighbour_values = _find_neighbour_values(array, k, j, i)
                ratios[k, j, i] = _calculate_max_ratio_from_neighbours(
                    value, neighbour_values
                )
    return ratios


@njit(cache=True)
def _find_neighbour_values(array: np.ndarray, k: int, j: int, i: int) -> np.ndarray:
    neighbour_values = np.zeros(6, dtype=float)
    neighbouring_indices = [
        [k, j, i + 1],
        [k, j, i - 1],
        [k, j + 1, i],
        [k, j - 1, i],
        [k + 1, j, i],
        [k - 1, j, i],
    ]
    for i, index in enumerate(neighbouring_indices):
        index_is_valid = True

        for index_i, dimension_i in zip(index, array.shape):
            if index_i < 0 or index_i >= dimension_i:
                index_is_valid = False
                break

        if index_is_valid:
            neighbour_values[i] = array[index[0], index[1], index[2]]

    return neighbour_values


@njit(cache=True)
def _calculate_max_ratio_from_neighbours(
    value: float, neighbour_values: np.ndarray
) -> float:
    ratios = np.divide(value, neighbour_values)

    # Invert the values that are less than 1
    ratios = np.where(ratios < 1, 1 / ratios, ratios)

    # Remove the infinities
    ratios = ratios[np.isfinite(ratios)]

    if ratios.size > 0:
        return np.max(ratios)
    else:
        return 1



"""
CLI for the ww_gvr package.
"""
import os
from copy import deepcopy
from enum import Enum
from pathlib import Path
from typing import Dict, Optional, Tuple

from f4enix.input.ww_gvr import WW

MAIN_MENU = """
 ***********************************************
        Weight window manipulator and GVR
 ***********************************************

 * Open weight window file   (open)   
 * Display ww information    (info)   
 * Write ww                  (write)
 * Export as VTK             (vtk)
 * Plot                      (plot)    
 * Weight window operation   (operate)
 * GVR generation            (gvr)
 * Exit                      (end)    
"""

OPERATE_MENU = """             
 * Softening and normalize   (soft)
 * Add                       (add)
 * Remove                    (rem)
 * Mitigate long histories   (mit)
 * Exit                      (end)
"""


class Command(Enum):
    OPEN = "open"
    INFO = "info"
    WRITE = "write"
    VTK = "vtk"
    PLOT = "plot"
    OPERATE = "operate"
    GVR = "gvr"
    END = "end"
    EMPTY = "empty"


class CommandOperate(Enum):
    SOFT = "soft"
    ADD = "add"
    REMOVE = "remove"
    MITIGATE = "mitigate"
    END = "end"
    EMPTY = "empty"


class Menu:
    def __init__(self) -> None:
        """
        Command-line interface for the ww_gvr package for handling weight windows (WW).

        Running Menu() will start the CLI loop directly.
        """
        self.weight_windows: Dict[str, WW] = {}
        self.extra_text: str = ""
        self.command_map = {
            Command.OPEN: self._handle_open,
            Command.INFO: self._handle_info,
            Command.WRITE: self._handle_write,
            Command.VTK: self._handle_vtk,
            Command.PLOT: self._handle_plot,
            Command.OPERATE: self._handle_operate,
            Command.GVR: self._handle_gvr,
            Command.END: exit,
        }
        self.command_operate_map = {
            CommandOperate.SOFT: self._handle_soft,
            CommandOperate.ADD: self._handle_add,
            CommandOperate.REMOVE: self._handle_remove,
            CommandOperate.MITIGATE: self._handle_mitigate,
            CommandOperate.END: self._display_menu,
        }
        self._main_menu_loop()

    def _main_menu_loop(self):
        while True:
            self._display_menu()
            self._get_command()

    def _display_menu(self):
        self.clear_screen()
        print(MAIN_MENU)
        print(self.extra_text)

    def _display_operate_menu(self):
        self.clear_screen()
        print(OPERATE_MENU)
        print(self.extra_text)

    def _get_command(self):
        while True:
            attempt = input(" Enter command: ").lower()
            try:
                function = self.command_map[Command(attempt)]
                function()
                break
            except ValueError:
                print(" Invalid command...")

    def _handle_open(self):
        file_path = Path(input(" Enter file path: "))

        if file_path.stem in self.weight_windows:
            print(" Weight window already loaded...")
            self._handle_open()

        try:
            weight_window = WW.load_from_ww_file(file_path)
            self.weight_windows[file_path.stem] = weight_window
            self.extra_text = f" {file_path.stem} loaded..."
        except FileNotFoundError:
            self.extra_text = " File not found..."

    def _handle_info(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = self.weight_windows[ww_key]
        self.extra_text = ww.info

    def _handle_write(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = self.weight_windows[ww_key]
        ww.write_to_ww_file()
        self.extra_text = f" {ww_key}_written saved!"

    def _handle_vtk(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = self.weight_windows[ww_key]
        ww.export_as_vtk()
        self.extra_text = f" {ww_key}.vtk saved!"

    def _handle_plot(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = self.weight_windows[ww_key]
        ww.geometry.plot()

    def _handle_gvr(self):
        file_path = Path(input(" Enter Meshtally file path: "))

        if file_path.stem in self.weight_windows:
            print(" Weight window already loaded with that file name...")
            self._handle_gvr()

        maximum_splitting, softening = self._ask_gvr_parameters()

        try:
            weight_window = WW.create_gvr_from_meshtally_file(
                file_path, maximum_splitting
            )
            weight_window.soften(softening)
            self.weight_windows[file_path.stem] = weight_window
            self.extra_text = f" {file_path.stem} loaded..."
        except FileNotFoundError:
            self.extra_text = " File not found..."

    def _select_ww_key(self) -> Optional[str]:
        if len(self.weight_windows) == 0:
            self.extra_text = " No weight windows loaded..."
            return None

        if len(self.weight_windows) == 1:
            return list(self.weight_windows.keys())[0]

        for i, ww_key in enumerate(self.weight_windows.keys()):
            print(f" {i+1}. {ww_key}")
        while True:
            ww_key = input(" Select weight window by number: ")
            if ww_key.isdigit():
                ww_key = int(ww_key)
                if ww_key in range(1, len(self.weight_windows) + 1):
                    return list(self.weight_windows.keys())[ww_key - 1]
            print(" Invalid weight window number...")

    def _ask_gvr_parameters(self) -> Tuple[float, float]:
        try:
            maximum_splitting = float(
                input(" Enter maximum splitting ratio (default 5): ")
            )
        except ValueError:
            maximum_splitting = 5
        try:
            softening = float(input(" Enter softening factor (default 1): "))
        except ValueError:
            softening = 1.0

        return maximum_splitting, softening

    def _handle_operate(self):
        self.extra_text = ""
        self._display_operate_menu()
        self._get_operate_command()

    def _get_operate_command(self):
        while True:
            attempt = input(" Enter command: ").lower()
            try:
                function = self.command_operate_map[CommandOperate(attempt)]
                function()
                break
            except ValueError:
                print(" Invalid command...")

    def _handle_soft(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = deepcopy(self.weight_windows[ww_key])
        softening = float(input(" Enter softening factor: "))
        ww.soften(softening)
        ww.file_path = ww.file_path.with_name(f"{ww_key}_softened")
        self.weight_windows[ww.file_path.stem] = ww
        self.extra_text = f" {ww.file_path.stem} created..."

    def _handle_add(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = deepcopy(self.weight_windows[ww_key])
        normalization = float(input(" Enter normalization factor: "))
        softening = float(input(" Enter softening factor: "))
        ww.add_particle(norm=normalization, soft=softening)
        ww.file_path = ww.file_path.with_name(f"{ww_key}_2_particles")
        self.weight_windows[ww.file_path.stem] = ww
        self.extra_text = f" {ww.file_path.stem} created..."

    def _handle_remove(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = deepcopy(self.weight_windows[ww_key])
        ww.remove_particle()
        ww.file_path = ww.file_path.with_name(f"{ww_key}_1_particle")
        self.weight_windows[ww.file_path.stem] = ww
        self.extra_text = f" {ww.file_path.stem} created..."

    def _handle_mitigate(self):
        ww_key = self._select_ww_key()
        if ww_key is None:
            return
        ww = deepcopy(self.weight_windows[ww_key])
        max_ratio = float(input(" Enter the maximum ratio allowed: "))
        ww.mitigate_long_histories(max_ratio)
        ww.file_path = ww.file_path.with_name(f"{ww_key}_mitigated")
        self.weight_windows[ww.file_path.stem] = ww
        self.extra_text = f" {ww.file_path.stem} created..."

    @staticmethod
    def clear_screen():
        """
        Clears the terminal screen for Windows or Linux.
        """
        if os.name == "nt":
            os.system("cls")
        else:
            os.system("clear")
